/**
 *  '$RCSfile: QuerySpecification.java,v $'
 *    Purpose: A Class that represents a structured query, and can be 
 *             constructed from an XML serialization conforming to 
 *             pathquery.dtd. The printSQL() method can be used to print 
 *             a SQL serialization of the query.
 *  Copyright: 2000 Regents of the University of California and the
 *             National Center for Ecological Analysis and Synthesis
 *    Authors: Matt Jones
 *    Release: @release@
 *
 *   '$Author: tao $'
 *     '$Date: 2003/03/05 18:18:02 $'
 * '$Revision: 1.46 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package edu.ucsb.nceas.metacat;

import edu.ucsb.nceas.dbadapter.*;

import java.io.*;
import java.util.Hashtable;
import java.util.Stack;
import java.util.Vector;
import java.util.Enumeration;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;
import org.xml.sax.helpers.DefaultHandler;

/**
 * A Class that represents a structured query, and can be 
 * constructed from an XML serialization conforming to @see pathquery.dtd. 
 * The printSQL() method can be used to print a SQL serialization of the query.
 */
public class QuerySpecification extends DefaultHandler {
 
  /** flag determining whether extended query terms are present */
  private boolean containsExtendedSQL=false;
  /** Identifier for this query document */
  private String meta_file_id;
  /** Title of this query */
  private String queryTitle;
  /** List of document types to be returned using package back tracing */
  private Vector returnDocList;
  /** List of document types to be searched */
  private Vector filterDocList;
  /** List of fields to be returned in result set */
  private Vector returnFieldList;
  /** List of users owning documents to be searched */
  private Vector ownerList;
  /** List of sites/scopes used to constrain search */
  private Vector siteList;
  /** The root query group that contains the recursive query constraints */
  private QueryGroup query = null;
 
  // Query data structures used temporarily during XML parsing
  private Stack elementStack;
  private Stack queryStack;
  private String currentValue;
  private String currentPathexpr;
  private String parserName = null;
  private String accNumberSeparator = null;
  private static final AbstractDatabase dbAdapter = MetaCatUtil.dbAdapter;
  
  private int countPercentageSearchItem = 0;
  private boolean percentageSearch = false;
  
  private String userName = null;
  private static final String PUBLIC = "public";
  private String [] group = null;

  public static final String ATTRIBUTESYMBOL = "@";
  private boolean hasAttributeReturnField = false;
  private Hashtable attributeReturnList = new Hashtable();
  private int countAttributeReturnField = 0;
  /**
   * construct an instance of the QuerySpecification class 
   *
   * @param queryspec the XML representation of the query (should conform
   *                  to pathquery.dtd) as a Reader
   * @param parserName the fully qualified name of a Java Class implementing
   *                  the org.xml.sax.XMLReader interface
   */
  public QuerySpecification( Reader queryspec, String parserName,
         String accNumberSeparator ) throws IOException {
    super();
    
    // Initialize the class variables
    returnDocList = new Vector();
    filterDocList = new Vector();
    elementStack = new Stack();
    queryStack   = new Stack();
    returnFieldList = new Vector();
    ownerList = new Vector();
    siteList = new Vector();
    this.parserName = parserName;
    this.accNumberSeparator = accNumberSeparator;

    // Initialize the parser and read the queryspec
    XMLReader parser = initializeParser();
    if (parser == null) {
      System.err.println("SAX parser not instantiated properly.");
    }
    try {
      parser.parse(new InputSource(queryspec));
    } catch (SAXException e) {
      System.err.println("error parsing data in " + 
                         "QuerySpecification.QuerySpecification");
      System.err.println(e.getMessage());
    }
  }

  /**
   * construct an instance of the QuerySpecification class 
   *
   * @param queryspec the XML representation of the query (should conform
   *                  to pathquery.dtd) as a String
   * @param parserName the fully qualified name of a Java Class implementing
   *                  the org.xml.sax.Parser interface
   */
  public QuerySpecification( String queryspec, String parserName,
         String accNumberSeparator) throws IOException {
    this(new StringReader(queryspec), parserName, accNumberSeparator);
  }
  
  /**
   * Method to set user name
   *
   * @param myName  the user name
   */
  public void setUserName(String myName)
  {
    this.userName = myName;
  }
  
  /**
   * Method to set user group
   *
   * @param myGroup  the user group
   */
  public void setGroup(String [] myGroup)
  {
    this.group = myGroup;
  }
  /**
   * Method to indicate this query is a percentage search
   */
  public boolean isPercentageSearch()
  {
    return percentageSearch;
  }
  /*
   * Method to get owner query. If it is owner it has all permission
   */
  private String createOwerQuery()
  {
    String ownerQuery = null;
    ownerQuery = "SELECT docid FROM xml_documents WHERE ";
    if (userName != null && !userName.equals(""))
    {
      ownerQuery = ownerQuery + "user_owner ='"+ userName +"'";
    }
    
    MetaCatUtil.debugMessage("OwnerQuery: "+ownerQuery, 30);
    return ownerQuery;
  }
  
  /*
   * Method to create query for xml_access, this part is to get docid list which
   * have a allow rule for a given user
   */
  private String createAllowRuleQuery()
  {
    String allowQuery = null;
    String allowString = constructAllowString();
    allowQuery ="SELECT docid from xml_access WHERE( "+allowString;
    allowQuery = allowQuery +") AND subtreeid IS NULL";
    MetaCatUtil.debugMessage("allow query is: "+ allowQuery, 30);
    return allowQuery;
    
  
  }
  
  /* Method to construct a allow rule string */
  private String constructAllowString()
  {
    String allowQuery ="";
     // add allow rule for user name
     if (userName != null && !userName.equals(""))
    {
      allowQuery = allowQuery +"(principal_name = '" + userName 
                              +"' AND perm_type = 'allow'"
                              +" AND (permission='4' OR permission='7'))";
    }
    // add allow rule for public
    allowQuery = allowQuery +"OR (principal_name = '" + PUBLIC 
                              +"' AND perm_type = 'allow'"
                              +" AND (permission='4' OR permission='7'))";
    
    // add allow rule for group
    if (group != null)
    {
      for (int i = 0; i< group.length; i++)
      {
        String groupUint = group[i];
        if (groupUint != null && !groupUint.equals(""))
        {
          allowQuery = allowQuery +" OR (principal_name = '" + groupUint 
                              +"' AND perm_type = 'allow'"
                              +" AND (permission='4' OR permission='7'))";
        }//if
      }//for
    }//if
    MetaCatUtil.debugMessage("allow string is: "+ allowQuery, 40);
    return allowQuery;
  }

   /*
   * Method to create query for xml_access, this part is to get docid list which
   * have a deny rule and perm_order is allowFirst for a given user. This means
   * the user will be denied to read
   */
  private String createDenyRuleQuery()
  {
    String denyQuery = null;
    String denyString = constructDenyString();
    denyQuery ="SELECT docid from xml_access WHERE( " + denyString;
    denyQuery = denyQuery + ") AND subtreeid IS NULL ";
    MetaCatUtil.debugMessage("denyquery is: "+ denyQuery, 30);
    return denyQuery;
  
  }
  /* Construct deny string */
  private String constructDenyString()
  {
    String denyQuery ="";
    // add deny rule for user name
    if (userName != null && !userName.equals(""))
    {
      denyQuery = denyQuery +"(principal_name = '" + userName 
                              +"' AND perm_type = 'deny' "
                              +"AND perm_order ='allowFirst'"
                              +" AND (permission='4' OR permission='7'))";
    }
    // add deny rule for public
    denyQuery = denyQuery +"OR (principal_name = '" + PUBLIC 
                               +"' AND perm_type = 'deny' "
                               +"AND perm_order ='allowFirst'"
                               +" AND (permission='4' OR permission='7'))";
    
    // add allow rule for group
    if (group != null)
    {
      for (int i = 0; i< group.length; i++)
      {
        String groupUint = group[i];
        if (groupUint != null && !groupUint.equals(""))
        {
          denyQuery = denyQuery +" OR (principal_name = '" + groupUint 
                                +"' AND perm_type = 'deny' "
                                +"AND perm_order ='allowFirst'"
                                +" AND (permission='4' OR permission='7'))";
        }//if
      }//for
    }//if
    return denyQuery;
  }
  
  /**
   * Method to append a access control query to SQL. So in DBQuery class, we can
   * get docid from both user specified query and access control query. We don't
   * need to checking permission after we get the doclist. It will be good to 
   * performance
   *
   */
  public String getAccessQuery()
  {
    String accessQuery = null;
    String onwer = createOwerQuery();
    String allow = createAllowRuleQuery();
    String deny = createDenyRuleQuery();
    accessQuery = " AND (docid IN("+ onwer + ")";
    accessQuery = accessQuery + " OR (docid IN (" + allow + ")" 
                 + " AND docid NOT IN ("+ deny + ")))";
    MetaCatUtil.debugMessage("accessquery is: "+ accessQuery, 30);
    return accessQuery;
  }
  
  /** Main routine for testing */
  static public void main(String[] args) {

     if (args.length < 1) {
       System.err.println("Wrong number of arguments!!!");
       System.err.println("USAGE: java QuerySpecification <xmlfile>");
       return;
     } else {
       int i = 0;
       boolean useXMLIndex = true;
       if ( args[i].equals( "-noindex" ) ) {
         useXMLIndex = false;
         i++;
       }
       String xmlfile  = args[i];

       try {
         MetaCatUtil util = new MetaCatUtil();
         FileReader xml = new FileReader(new File(xmlfile));
         QuerySpecification qspec = 
                 new QuerySpecification(xml, util.getOption("saxparser"),
                                        util.getOption("accNumberSeparator"));
         System.out.println(qspec.printSQL(useXMLIndex));

       } catch (IOException e) {
         System.err.println(e.getMessage());
       }
         
     }
  }
  
  /**
   * Returns true if the parsed query contains and extended xml query 
   * (i.e. there is at least one &lt;returnfield&gt; in the pathquery document)
   */
  public boolean containsExtendedSQL()
  {
    if(containsExtendedSQL)
    {
      return true;
    }
    else
    {
      return false;
    }
  }
  
  /**
   * A method to get if the query has an attribute return field
   */
  public boolean containAttributeReturnField()
  {
    return hasAttributeReturnField;
  }
  
  /**
   * Accessor method to return the identifier of this Query
   */
  public String getIdentifier()
  {
    return meta_file_id;
  }

  /**
   * method to set the identifier of this query
   */
  public void setIdentifier(String id) {
    this.meta_file_id = id;
  }

  /**
   * Accessor method to return the title of this Query
   */
  public String getQueryTitle()
  {
    return queryTitle;
  }

  /**
   * method to set the title of this query
   */
  public void setQueryTitle(String title)
  {
    this.queryTitle = title;
  }

  /**
   * Accessor method to return a vector of the return document types as
   * defined in the &lt;returndoctype&gt; tag in the pathquery dtd.
   */
  public Vector getReturnDocList()
  {
    return this.returnDocList;
  }

  /**
   * method to set the list of return docs of this query
   */
  public void setReturnDocList(Vector returnDocList)
  {
    this.returnDocList = returnDocList;
  }

  /**
   * Accessor method to return a vector of the filter doc types as
   * defined in the &lt;filterdoctype&gt; tag in the pathquery dtd.
   */
  public Vector getFilterDocList()
  {
    return this.filterDocList;
  }

  /**
   * method to set the list of filter docs of this query
   */
  public void setFilterDocList(Vector filterDocList)
  {
    this.filterDocList = filterDocList;
  }

  /**
   * Accessor method to return a vector of the extended return fields as
   * defined in the &lt;returnfield&gt; tag in the pathquery dtd.
   */
  public Vector getReturnFieldList()
  {
    return this.returnFieldList; 
  }

  /**
   * method to set the list of fields to be returned by this query
   */
  public void setReturnFieldList(Vector returnFieldList)
  {
    this.returnFieldList = returnFieldList;
  }

  /**
   * Accessor method to return a vector of the owner fields as
   * defined in the &lt;owner&gt; tag in the pathquery dtd.
   */
  public Vector getOwnerList()
  {
    return this.ownerList;
  }

  /**
   * method to set the list of owners used to constrain this query
   */
  public void setOwnerList(Vector ownerList)
  {
    this.ownerList = ownerList;
  }

  /**
   * Accessor method to return a vector of the site fields as
   * defined in the &lt;site&gt; tag in the pathquery dtd.
   */
  public Vector getSiteList()
  {
    return this.siteList;
  }

  /**
   * method to set the list of sites used to constrain this query
   */
  public void setSiteList(Vector siteList)
  {
    this.siteList = siteList;
  }

  /**
   * get the QueryGroup used to express query constraints
   */
  public QueryGroup getQueryGroup()
  {
    return query;
  }

  /**
   * Set up the SAX parser for reading the XML serialized query
   */
  private XMLReader initializeParser() {
    XMLReader parser = null;

    // Set up the SAX document handlers for parsing
    try {

      // Get an instance of the parser
      parser = XMLReaderFactory.createXMLReader(parserName);

      // Set the ContentHandler to this instance
      parser.setContentHandler(this);

      // Set the error Handler to this instance
      parser.setErrorHandler(this);

    } catch (Exception e) {
       System.err.println("Error in QuerySpcecification.initializeParser " + 
                           e.toString());
    }

    return parser;
  }

  /**
   * callback method used by the SAX Parser when the start tag of an 
   * element is detected. Used in this context to parse and store
   * the query information in class variables.
   */
  public void startElement (String uri, String localName, 
                            String qName, Attributes atts) 
         throws SAXException {
    BasicNode currentNode = new BasicNode(localName);
    // add attributes to BasicNode here
    if (atts != null) {
      int len = atts.getLength();
      for (int i = 0; i < len; i++) {
        currentNode.setAttribute(atts.getLocalName(i), atts.getValue(i));
      }
    }

    elementStack.push(currentNode); 
    if (currentNode.getTagName().equals("querygroup")) {
      QueryGroup currentGroup = new QueryGroup(
                                currentNode.getAttribute("operator"));
      if (query == null) {
        query = currentGroup;
      } else {
        QueryGroup parentGroup = (QueryGroup)queryStack.peek();
        parentGroup.addChild(currentGroup);
      }
      queryStack.push(currentGroup);
    }
  }

  /**
   * callback method used by the SAX Parser when the end tag of an 
   * element is detected. Used in this context to parse and store
   * the query information in class variables.
   */
  public void endElement (String uri, String localName,
                          String qName) throws SAXException {
    BasicNode leaving = (BasicNode)elementStack.pop(); 
    if (leaving.getTagName().equals("queryterm")) {
      boolean isCaseSensitive = (new Boolean(
              leaving.getAttribute("casesensitive"))).booleanValue();
      QueryTerm currentTerm = null;
      if (currentPathexpr == null) {
        currentTerm = new QueryTerm(isCaseSensitive,
                      leaving.getAttribute("searchmode"),currentValue);
      } else {
        currentTerm = new QueryTerm(isCaseSensitive,
                      leaving.getAttribute("searchmode"),currentValue,
                      currentPathexpr);
      }
      QueryGroup currentGroup = (QueryGroup)queryStack.peek();
      currentGroup.addChild(currentTerm);
      currentValue = null;
      currentPathexpr = null;
    } else if (leaving.getTagName().equals("querygroup")) {
      QueryGroup leavingGroup = (QueryGroup)queryStack.pop();
    }
  }

  /**
   * callback method used by the SAX Parser when the text sequences of an 
   * xml stream are detected. Used in this context to parse and store
   * the query information in class variables.
   */
  public void characters(char ch[], int start, int length) {

    String inputString = new String(ch, start, length);
    BasicNode currentNode = (BasicNode)elementStack.peek(); 
    String currentTag = currentNode.getTagName();
    if (currentTag.equals("meta_file_id")) {
      meta_file_id = inputString;
    } else if (currentTag.equals("querytitle")) {
      queryTitle = inputString;
    } else if (currentTag.equals("value")) {
      currentValue = inputString;
    } else if (currentTag.equals("pathexpr")) {
      currentPathexpr = inputString;
    } else if (currentTag.equals("returndoctype")) {
      returnDocList.add(inputString);
    } else if (currentTag.equals("filterdoctype")) {
      filterDocList.add(inputString);
    } else if (currentTag.equals("returnfield")) {
      // make sure if return fields has an attribute or not
      if (inputString.indexOf(ATTRIBUTESYMBOL) ==-1)
      {
        // no attribute value will be returned
        returnFieldList.add(inputString);
        containsExtendedSQL = true;
      }
      else
      {
        // has a attribute return field
        // divied the return filed into two parts, one is path and the
        // other is attribue name
        String returnPath = newPathExpressionWithOutAttribute(inputString);
        String attributeName = getAttributeName(inputString);
        Vector pathInfo = new Vector();
        // the vector has the information about return path and attributename
        pathInfo.addElement(returnPath);
        pathInfo.addElement(attributeName);
        // put the vector into a hash table. The reseaon why don't put
        // return path or attributename as a key is because they are not unique
        attributeReturnList.put
                            (new Integer(countAttributeReturnField), pathInfo);
        countAttributeReturnField++;
        hasAttributeReturnField = true;
        containsExtendedSQL = true;
        
      }
    } else if (currentTag.equals("filterdoctype")) {
      filterDocList.add(inputString);
    } else if (currentTag.equals("owner")) {
      ownerList.add(inputString);
    } else if (currentTag.equals("site")) {
      siteList.add(inputString);
    }
  }

  /**
   * create a SQL serialization of the query that this instance represents
   */
  public String printSQL(boolean useXMLIndex) {
    
   
    StringBuffer self = new StringBuffer();

    self.append("SELECT docid,docname,doctype,");
    self.append("date_created, date_updated, rev ");
    self.append("FROM xml_documents WHERE docid IN (");

    // This determines the documents that meet the query conditions
    self.append(query.printSQL(useXMLIndex));

    self.append(") ");
 
    // Add SQL to filter for doctypes requested in the query
    // This is an implicit OR for the list of doctypes. Only doctypes in this
    // list will be searched if the tag is present
    if (!filterDocList.isEmpty()) {
      boolean firstdoctype = true;
      self.append(" AND ("); 
      Enumeration en = filterDocList.elements();
      while (en.hasMoreElements()) {
        String currentDoctype = (String)en.nextElement();
        if (firstdoctype) {
           firstdoctype = false;
           self.append(" doctype = '" + currentDoctype + "'"); 
        } else {
          self.append(" OR doctype = '" + currentDoctype + "'"); 
        }
      }
      self.append(") ");
    }

    // Add SQL to filter for owners requested in the query
    // This is an implicit OR for the list of owners
    if (!ownerList.isEmpty()) {
      boolean first = true;
      self.append(" AND ("); 
      Enumeration en = ownerList.elements();
      while (en.hasMoreElements()) {
        String current = (String)en.nextElement();
        if (first) {
           first = false;
           self.append(" user_owner = '" + current + "'"); 
        } else {
          self.append(" OR user_owner = '" + current + "'"); 
        }
      }
      self.append(") ");
    }

    // Add SQL to filter for sites requested in the query
    // This is an implicit OR for the list of sites
    if (!siteList.isEmpty()) {
      boolean first = true;
      self.append(" AND ("); 
      Enumeration en = siteList.elements();
      while (en.hasMoreElements()) {
        String current = (String)en.nextElement();
        if (first) {
           first = false;
           self.append(" SUBSTR(docid, 1, INSTR(docid, '" +
               accNumberSeparator + "')-1) = '" + current + "'"); 
        } else {
          self.append(" OR SUBSTR(docid, 1, INSTR(docid, '" +
               accNumberSeparator + "')-1) = '" + current + "'"); 
        }
      }
      self.append(") ");
    }
    
    // if there is only one percentage search item, this query is a percentage
    // search query
    if (countPercentageSearchItem ==1)
    {
      
      percentageSearch =true;
    }
    
    return self.toString();
  }
  
  /** This sql command will selecet startnodeid and endnodeid that user can NOT
    * access
    */
  public String printAccessControlSQLForReturnField(String doclist)
  {
    StringBuffer sql = new StringBuffer();
    String allowString = constructAllowString();
    String denyString = constructDenyString();
    sql.append("SELECT distinct startnodeid, endnodeid from xml_access ");
    sql.append("WHERE docid in (");
    sql.append(doclist);
    sql.append(") AND startnodeid IS NOT NULL AND ");
    sql.append("(");
    sql.append("(");
    sql.append("startnodeid NOT IN (SELECT startnodeid from xml_access, xml_documents ");
    sql.append(" WHERE xml_access.docid = xml_documents.docid");
    sql.append(" AND xml_documents.user_owner ='");
    sql.append(userName);
    sql.append("' AND xml_access.startnodeid IS NOT NULL)"); 
    sql.append(")");
    sql.append(" AND ");
    sql.append("(");
    sql.append("(startnodeid NOT IN (SELECT startnodeid from xml_access where( ");
    sql.append(allowString);
    sql.append(") AND (startnodeid IS NOT NULL))");
    sql.append(")");
    sql.append(" OR (startnodeid IN (SELECT startnodeid from xml_access where( ");
    sql.append(denyString);
    sql.append(") AND (startnodeid IS NOT NULL))");
    sql.append(")");
    sql.append(")");
    sql.append(")");
    MetaCatUtil.debugMessage("accessControlSQLForReturnField: " +
                             sql.toString(), 30);
    return sql.toString();
  }
  
  /**
   * This method prints sql based upon the &lt;returnfield&gt; tag in the
   * pathquery document.  This allows for customization of the 
   * returned fields
   * @param doclist the list of document ids to search by
   * @param unaccessableNodePair  the node pair(start id and end id) which this
   *                               user could not access it
   */
  public String printExtendedSQL(String doclist, Hashtable unaccessableNodePair)
  {  
    StringBuffer self = new StringBuffer();
    self.append("select xml_nodes.docid, xml_index.path, xml_nodes.nodedata, ");
    self.append("xml_nodes.parentnodeid ");
    self.append("from xml_index, xml_nodes where xml_index.nodeid=");
    self.append("xml_nodes.parentnodeid and (xml_index.path like '");
    boolean firstfield = true;
    //put the returnfields into the query
    //the for loop allows for multiple fields
    for(int i=0; i<returnFieldList.size(); i++)
    {
      if(firstfield)
      {
        firstfield = false;
        self.append((String)returnFieldList.elementAt(i));
        self.append("' ");
      }
      else
      {
        self.append("or xml_index.path like '");
        self.append((String)returnFieldList.elementAt(i));
        self.append("' ");
      }
    }
    self.append(") AND xml_nodes.docid in (");
    //self.append(query.printSQL());
    self.append(doclist);
    self.append(")");
    self.append(" AND xml_nodes.nodetype = 'TEXT'");
    
    // add control part for extended query
    Enumeration en = unaccessableNodePair.keys();
    
    while (en.hasMoreElements())
    {
      // Get control pairs in object
      Long startNodeIdObject = (Long)en.nextElement();
      Long endNodeIdObject = (Long)unaccessableNodePair.get(startNodeIdObject);
      // change it to long
      long startNodeId = startNodeIdObject.longValue();
      long endNodeId   = endNodeIdObject.longValue();
      // add into query
      self.append(" AND( xml_nodes.nodeid < ");
      self.append(startNodeId);
      self.append(" OR xml_nodes.nodeid > ");
      self.append(endNodeId);
      self.append(")");
    }

   
    return self.toString();
  }
  
  /**
   * This method prints sql based upon the returnfield tag in the
   * pathquery document has an attribute.  This allows for customization of the 
   * returned fields
   * @param doclist the list of document ids to search by
   */
  public String printAttributeQuery(String doclist)
  {
    StringBuffer self = new StringBuffer();
    self.append("select xml_nodes.docid, xml_index.path, ");
    self.append("xml_nodes.nodedata, xml_nodes.nodename ");
    self.append("from xml_index, xml_nodes where xml_index.nodeid=");
    self.append("xml_nodes.parentnodeid and (");
    boolean firstfield = true;
    //put the returnfields attributes into the query
    //the for loop allows for multiple fields and attributes
    Enumeration returnAttributes = attributeReturnList.elements();
    while (returnAttributes.hasMoreElements())
    {
      Vector currentVector = (Vector)returnAttributes.nextElement();
      String returnPath = (String)currentVector.elementAt(0);
      String attributeName = (String)currentVector.elementAt(1);
      if(firstfield)
      {
        firstfield = false;
        self.append("( xml_index.path like '");
        self.append(returnPath);
        self.append("' AND xml_nodes.nodename like '");
        self.append(attributeName);
        self.append("') ");
      }
      else
      {
        self.append(" or ( xml_index.path like '");
        self.append(returnPath);
        self.append("' AND xml_nodes.nodename like '");
        self.append(attributeName);
        self.append("') "); 
      }
    }
    self.append(") AND xml_nodes.docid in (");
    //self.append(query.printSQL());
    self.append(doclist);
    self.append(")");
    self.append(" AND xml_nodes.nodetype = 'ATTRIBUTE'");
    MetaCatUtil.debugMessage("Attribute query: "+self.toString(), 30);
   
    return self.toString();
  }
  
  public static String printRelationSQL(String docid)
  {
    StringBuffer self = new StringBuffer();
    self.append("select subject, relationship, object, subdoctype, ");
    self.append("objdoctype from xml_relation ");
    self.append("where docid like '").append(docid).append("'");
    return self.toString();
  }
   
  /**
   * Prints sql that returns all relations in the database.
   */
  public static String printPackageSQL()
  {
    StringBuffer self = new StringBuffer();
    self.append("select z.nodedata, x.nodedata, y.nodedata from ");
    self.append("(select nodeid, parentnodeid from xml_index where path like ");
    self.append("'triple/subject') s, (select nodeid, parentnodeid ");
    self.append("from xml_index where path like ");
    self.append("'triple/relationship') rel, ");
    self.append("(select nodeid, parentnodeid from xml_index where path like ");
    self.append("'triple/object') o, ");
    self.append("xml_nodes x, xml_nodes y, xml_nodes z ");
    self.append("where s.parentnodeid = rel.parentnodeid ");
    self.append("and rel.parentnodeid = o.parentnodeid ");
    self.append("and x.parentnodeid in (rel.nodeid) ");
    self.append("and y.parentnodeid in (o.nodeid) ");
    self.append("and z.parentnodeid in (s.nodeid) ");
    //self.append("and z.nodedata like '%");
    //self.append(docid);
    //self.append("%'");
    return self.toString();
  }
  
  /**
   * Prints sql that returns all relations in the database that were input
   * under a specific docid
   * @param docid the docid to search for.
   */
  public static String printPackageSQL(String docid)
  {
    StringBuffer self = new StringBuffer();
    self.append("select z.nodedata, z.parentnodeid, ");
    self.append("x.nodedata, x.parentnodeid, y.nodedata, y.parentnodeid from ");
    self.append("(select nodeid, parentnodeid from xml_index where path like ");
    self.append("'triple/subject') s, (select nodeid, parentnodeid ");
    self.append("from xml_index where path like ");
    self.append("'triple/relationship') rel, ");
    self.append("(select nodeid, parentnodeid from xml_index where path like ");
    self.append("'triple/object') o, ");
    self.append("xml_nodes x, xml_nodes y, xml_nodes z ");
    self.append("where s.parentnodeid = rel.parentnodeid ");
    self.append("and rel.parentnodeid = o.parentnodeid ");
    self.append("and x.parentnodeid in (rel.nodeid) ");
    self.append("and y.parentnodeid in (o.nodeid) ");
    self.append("and z.parentnodeid in (s.nodeid) ");
    self.append("and z.docid like '").append(docid).append("'");
    
    return self.toString();
  }
  
  /**
   * Returns all of the relations that has a certain docid in the subject
   * or the object.
   * 
   * @param docid the docid to search for
   */
  public static String printPackageSQL(String subDocidURL, String objDocidURL)
  {
    StringBuffer self = new StringBuffer();
    self.append("select z.nodedata, x.nodedata, y.nodedata from ");
    self.append("(select nodeid, parentnodeid from xml_index where path like ");
    self.append("'triple/subject') s, (select nodeid, parentnodeid ");
    self.append("from xml_index where path like ");
    self.append("'triple/relationship') rel, ");
    self.append("(select nodeid, parentnodeid from xml_index where path like ");
    self.append("'triple/object') o, ");
    self.append("xml_nodes x, xml_nodes y, xml_nodes z ");
    self.append("where s.parentnodeid = rel.parentnodeid ");
    self.append("and rel.parentnodeid = o.parentnodeid ");
    self.append("and x.parentnodeid in (rel.nodeid) ");
    self.append("and y.parentnodeid in (o.nodeid) ");
    self.append("and z.parentnodeid in (s.nodeid) ");
    self.append("and (z.nodedata like '");
    self.append(subDocidURL);
    self.append("' or y.nodedata like '");
    self.append(objDocidURL);
    self.append("')");
    return self.toString();
  }
  
  public static String printGetDocByDoctypeSQL(String docid)
  {
    StringBuffer self = new StringBuffer();

    self.append("SELECT docid,docname,doctype,");
    self.append("date_created, date_updated ");
    self.append("FROM xml_documents WHERE docid IN (");
    self.append(docid).append(")");
    return self.toString();
  }
  
  /**
   * create a String description of the query that this instance represents.
   * This should become a way to get the XML serialization of the query.
   */
  public String toString() {
    return "meta_file_id=" + meta_file_id + "\n" + query;
//DOCTITLE attr cleared from the db
//    return "meta_file_id=" + meta_file_id + "\n" + 
//           "querytitle=" + querytitle + "\n" + query;
  }
  
  /* A method to get rid of attribute part in path expression*/
  public static String newPathExpressionWithOutAttribute(String pathExpression)
  {
      if (pathExpression == null)
      {
        return null;
      }
      int index = pathExpression.lastIndexOf(ATTRIBUTESYMBOL);
      String newExpression = null;
      if (index != 1)
      {
        newExpression=pathExpression.substring(0, index-1);
      } 
      MetaCatUtil.debugMessage("The path expression without attributes: )" + 
                               newExpression, 30);
      return newExpression;
  }
    
  /* A method to get attribute name from path */
  public static String getAttributeName(String path)
  {
      if (path == null)
      {
        return null;
      }
      int index = path.lastIndexOf(ATTRIBUTESYMBOL);
      int size = path.length();
      String attributeName = null;
       if (index != 1)
      {
        attributeName = path.substring(index+1, size);
      } 
      MetaCatUtil.debugMessage("The attirbute name from path: )" + 
                               attributeName, 30);
      return attributeName;
  }

  /** a utility class that represents a group of terms in a query */
  private class QueryGroup {
    private String operator = null;  // indicates how query terms are combined
    private Vector children = null;  // the list of query terms and groups

    /** 
     * construct a new QueryGroup 
     *
     * @param operator the boolean conector used to connect query terms 
     *                    in this query group
     */
    public QueryGroup(String operator) {
      this.operator = operator;
      children = new Vector();
    }

    /** 
     * Add a child QueryGroup to this QueryGroup
     *
     * @param qgroup the query group to be added to the list of terms
     */
    public void addChild(QueryGroup qgroup) {
      children.add((Object)qgroup); 
    }

    /**
     * Add a child QueryTerm to this QueryGroup
     *
     * @param qterm the query term to be added to the list of terms
     */
    public void addChild(QueryTerm qterm) {
      children.add((Object)qterm); 
    }

    /**
     * Retrieve an Enumeration of query terms for this QueryGroup
     */
    public Enumeration getChildren() {
      return children.elements();
    }
   
    /**
     * create a SQL serialization of the query that this instance represents
     */
    public String printSQL(boolean useXMLIndex) {
      StringBuffer self = new StringBuffer();
      boolean first = true;

      self.append("(");

      Enumeration en= getChildren();
      while (en.hasMoreElements()) {
        Object qobject = en.nextElement();
        if (first) {
          first = false;
        } else {
          self.append(" " + operator + " ");
        }
        if (qobject instanceof QueryGroup) {
          QueryGroup qg = (QueryGroup)qobject;
          self.append(qg.printSQL(useXMLIndex));
        } else if (qobject instanceof QueryTerm) {
          QueryTerm qt = (QueryTerm)qobject;
          self.append(qt.printSQL(useXMLIndex));
        } else {
          System.err.println("qobject wrong type: fatal error");
        }
      }
      self.append(") \n");
      return self.toString();
    }

    /**
     * create a String description of the query that this instance represents.
     * This should become a way to get the XML serialization of the query.
     */
    public String toString() {
      StringBuffer self = new StringBuffer();

      self.append("  (Query group operator=" + operator + "\n");
      Enumeration en= getChildren();
      while (en.hasMoreElements()) {
        Object qobject = en.nextElement();
        self.append(qobject);
      }
      self.append("  )\n");
      return self.toString();
    }
  }

  /** a utility class that represents a single term in a query */
  private class QueryTerm {
    private boolean casesensitive = false;
    private String searchmode = null;
    private String value = null;
    private String pathexpr = null;
    private boolean percentageSymbol = false;
   

    /**
     * Construct a new instance of a query term for a free text search
     * (using the value only)
     *
     * @param casesensitive flag indicating whether case is used to match
     * @param searchmode determines what kind of substring match is performed
     *        (one of starts-with|ends-with|contains|matches-exactly)
     * @param value the text value to match
     */
    public QueryTerm(boolean casesensitive, String searchmode, 
                     String value) {
      this.casesensitive = casesensitive;
      this.searchmode = searchmode;
      this.value = value;
    }

    /**
     * Construct a new instance of a query term for a structured search
     * (matching the value only for those nodes in the pathexpr)
     *
     * @param casesensitive flag indicating whether case is used to match
     * @param searchmode determines what kind of substring match is performed
     *        (one of starts-with|ends-with|contains|matches-exactly)
     * @param value the text value to match
     * @param pathexpr the hierarchical path to the nodes to be searched
     */
    public QueryTerm(boolean casesensitive, String searchmode, 
                     String value, String pathexpr) {
      this(casesensitive, searchmode, value);
      this.pathexpr = pathexpr;
    }

    /** determine if the QueryTerm is case sensitive */
    public boolean isCaseSensitive() {
      return casesensitive;
    }

    /** get the searchmode parameter */
    public String getSearchMode() {
      return searchmode;
    }
 
    /** get the Value parameter */
    public String getValue() {
      return value;
    }

    /** get the path expression parameter */
    public String getPathExpression() {
      return pathexpr;
    }

    /**
     * create a SQL serialization of the query that this instance represents
     */
    public String printSQL(boolean useXMLIndex) {
      StringBuffer self = new StringBuffer();

      // Uppercase the search string if case match is not important
      String casevalue = null;
      String nodedataterm = null;

      if (casesensitive) {
        nodedataterm = "nodedata";
        casevalue = value;
      } else {
        nodedataterm = "UPPER(nodedata)";
        casevalue = value.toUpperCase();
      }

      // Add appropriate wildcards to search string
      //String searchvalue = null;
      String searchexpr = null;
      if (searchmode.equals("starts-with")) {
        //searchvalue = casevalue + "%";
        searchexpr = nodedataterm + " LIKE '" + casevalue + "%' ";
      } else if (searchmode.equals("ends-with")) {
        //searchvalue = "%" + casevalue;
        searchexpr = nodedataterm + " LIKE '%" + casevalue + "' ";
      } else if (searchmode.equals("contains")) {
        //searchvalue = "%" + casevalue + "%";
        if (!casevalue.equals("%"))
        {
          searchexpr = nodedataterm + " LIKE '%" + casevalue + "%' ";
        }
        else
        {
          searchexpr = nodedataterm + " LIKE '" + casevalue + "' ";
          // find percentage symbol
          percentageSymbol = true;
        }
      } else if (searchmode.equals("equals")) {
        //searchvalue = casevalue;
        searchexpr = nodedataterm + " = '" + casevalue + "' ";
      } else if (searchmode.equals("isnot-equal")) {
        //searchvalue = casevalue;
        searchexpr = nodedataterm + " != '" + casevalue + "' ";
      } else { 
        //searchvalue = casevalue;
        String oper = null;
        if (searchmode.equals("greater-than")) {
          oper = ">";
        } else if (searchmode.equals("greater-than-equals")) {
          oper = ">=";
        } else if (searchmode.equals("less-than")) {
          oper = "<";
        } else if (searchmode.equals("less-than-equals")) {
          oper = "<=";
        } else {
          System.out.println("NOT expected case. NOT recognized operator: " +
                             searchmode);
          return null;
        }
        try {
          // it is number; numeric comparison
          // but we need to make sure there is no string in node data
          String getRidOfString = " AND UPPER(nodedata) = LOWER(nodedata)" +
                                  " AND LTRIM(nodedata) != ' ' " +
                                  " AND nodedata IS NOT NULL ";
          searchexpr = nodedataterm + " " + oper + " " +
                       new Double(casevalue) + " "+getRidOfString;          
        } catch (NumberFormatException nfe) {
          // these are characters; character comparison
          searchexpr = nodedataterm + " " + oper + " '" + casevalue + "' ";
        }
      }

      self.append("SELECT DISTINCT docid FROM xml_nodes WHERE \n");
      //self.append(nodedataterm + " LIKE " + "'" + searchvalue + "' ");
      self.append(searchexpr);

      if (pathexpr != null) 
      {
        
        // use XML Index
        if ( useXMLIndex ) 
        {
          if (!hasAttributeInPath(pathexpr))
          {
            // without attributes in path
            self.append("AND parentnodeid IN ");
          }
          else
          {
            // has a attribute in path
            String attributeName = getAttributeName(pathexpr);
            self.append("AND nodetype LIKE 'ATTRIBUTE' AND nodename LIKE '"+
                        attributeName + "' ");
            self.append("AND parentnodeid IN ");
            pathexpr = newPathExpressionWithOutAttribute(pathexpr);
            
          } 
          self.append("(SELECT nodeid FROM xml_index WHERE path LIKE " + 
                      "'" +  pathexpr + "') " );
        } 
        else 
        {
          // without using XML Index; using nested statements instead
          self.append("AND parentnodeid IN ");
          self.append(useNestedStatements(pathexpr));
        }
      }
      else
      {
        //if pathexpr is null and search value is %, is a percentageSearchItem
        // the count number will be increase one
        countPercentageSearchItem++;
        
       }

      return self.toString();
    }
    
    /* A method to judge if a path have attribute */
    private boolean hasAttributeInPath(String path)
    {
      if (path.indexOf(ATTRIBUTESYMBOL)!=-1)
      {
        return true;
      }
      else
      {
        return false;
      }
    }
    
   
    
    /* 
     * Constraint the query with @pathexp without using the XML Index,
     * but nested SQL statements instead. The query migth be slower.
     */
    private String useNestedStatements(String pathexpr)
    {
      StringBuffer nestedStmts = new StringBuffer();
      Vector nodes = new Vector();
      String path = pathexpr;
      int inx = 0;

      do {
        inx = path.lastIndexOf("/");

        nodes.addElement(path.substring(inx+1));
        path = path.substring(0, Math.abs(inx));
      } while ( inx > 0 );
      
      // nested statements
      int i = 0;
      for (i = 0; i < nodes.size()-1; i++) {
        nestedStmts.append("(SELECT nodeid FROM xml_nodes" + 
                           " WHERE nodename LIKE '" +
                             (String)nodes.elementAt(i) + "'" +
                           " AND parentnodeid IN ");
      }
      // for the last statement: it is without " AND parentnodeid IN "
      nestedStmts.append("(SELECT nodeid FROM xml_nodes" + 
                         " WHERE nodename LIKE '" +
                         (String)nodes.elementAt(i) + "'" );
      // node.size() number of closing brackets
      for (i = 0; i < nodes.size(); i++) {
        nestedStmts.append(")");
      }



      return nestedStmts.toString();
    }

    /**
     * create a String description of the query that this instance represents.
     * This should become a way to get the XML serialization of the query.
     */
    public String toString() {

      return this.printSQL(true);
    }
  }
}
